-- Project: Fractal for Spartan-3
-- Author : Matthias Alles
-- Mail   : alles(at) rhrk.uni-kk.de
--
-- This is the top-module that builds the connectivity of all used components and controls these.
-- clk_pad was 49,152 MHz
-- clk_2, the pixel clock, is 24,576 MHz
-- if you want to use different clocks you have to adapt the timing constants in the vga_timing
-- component to get a 640*480*60Hz timing. Also the initialization of the mouse might need some
-- adaptions (read the comments in the component).

library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use ieee.numeric_std.all;

entity frac_controller is
port (
	clk_pad  : in    std_logic;
	rst      : in    std_logic;
	ms_clk   : inout std_logic;
	ms_data  : inout std_logic;
	hsync    : out   std_logic;
	vsync    : out   std_logic;
	rgb      : out   std_logic_vector(14 downto 0));
end frac_controller;

architecture rtl of frac_controller is

	component vga_timing port(
		clk   : in std_logic;
		rst   : in std_logic;
		hsync : out std_logic;
		vsync : out std_logic;
		blank : out std_logic;
		x     : out std_logic_vector(9 downto 0);
		y     : out std_logic_vector(9 downto 0));
	end component;

	signal x : std_logic_vector(9 downto 0);
	signal y : std_logic_vector(9 downto 0);
	signal blank : std_logic;
	signal clk_2  : std_logic;

	component pixel_calc port(
		clk    : in std_logic;
		rst    : in std_logic;
		start  : in std_logic;
		c_x    : in signed(17 downto 0);
		c_y    : in signed(17 downto 0);
		result : out unsigned(4 downto 0));
	end component;

	component mouse port(
		clk          : in  std_logic;
		rst          : in  std_logic;
		interrupt    : out std_logic;
		data         : out std_logic_vector(31 downto 0);
		mouse_clk_I  : in  std_logic;
		mouse_clk_O  : out std_logic;
		mouse_clk_T  : out std_logic;
		mouse_data_I : in  std_logic;
		mouse_data_O : out std_logic;
		mouse_data_T : out std_logic);
	end component;

--	component IOBUF is
--	port (
--		I : in std_logic;
--		IO : inout std_logic;
--		O : out std_logic;
--		T : in std_logic
--	);
--	end component;

	component BUFG_CLKDV_SUBM is
	port(
		CLK_IN : in std_logic;
		RST    : in std_logic;
		CLK1X  : out std_logic;
		CLK_DIV  : out std_logic;
		LOCK   : out std_logic
	);
	end component;

	signal lock   : std_logic;
	signal clk    : std_logic;

	signal start : std_logic;
	type c_vector is array(0 to 7) of signed(17 downto 0);
	signal c_x : c_vector;
	signal c_y : signed(17 downto 0);
	type result_vector is array(0 to 7) of unsigned(4 downto 0);
	signal result : result_vector;
	signal result_reg : result_vector;

	signal y_offset : signed(17 downto 0);
	signal x_offset : c_vector;
	signal hsync_int : std_logic;
	signal vsync_int : std_logic;

	signal hsync_int_t : std_logic;
	signal vsync_int_t : std_logic;

	constant START_X : integer := -111072/4;
	constant START_Y : integer := 65536/3;

	signal clock_counter : unsigned(4 downto 0);
	type master_ctrl_type is (IDLE, RUNNING);
	signal master_ctrl_state: master_ctrl_type;
	
	signal index : integer range 0 to 7;
	signal state_zoom_auto : std_logic;
	signal zoom_in : std_logic;

	signal start_x_sig : signed(17 downto 0);
	signal start_y_sig : signed(17 downto 0);

	signal blank_delay: std_logic_vector(17 downto 0);

	-- mouse handling
	signal mouse_interrupt : std_logic;
	signal mouse_data      : std_logic_vector(31 downto 0);
	signal mouse_clk_I     : std_logic;
	signal mouse_clk_O     : std_logic;
	signal mouse_clk_T     : std_logic;
	signal mouse_data_I    : std_logic;
	signal mouse_data_O    : std_logic;
	signal mouse_data_T    : std_logic;
	
	function color_lut(input : in unsigned(4 downto 0)) return std_logic_vector is
	begin
		case to_integer(input) is
			when 0 => return "00000" & "00000" & "00000";
			when 1 => return "00010" & "00001" & "01011";
			when 2 => return "00011" & "00010" & "10001";
			when 3 => return "01001" & "00001" & "01110";
			when 4 => return "01110" & "00001" & "01010";
			when 5 => return "10001" & "00001" & "01000";
			when 6 => return "11000" & "00001" & "00011";
			when 7 => return "11110" & "00100" & "00001";
			when 8 => return "11111" & "01111" & "00000";
			when 9 => return "11111" & "10101" & "00010";
			when 10 => return "11110" & "11010" & "00100";
			when 11 => return "11110" & "11110" & "00101";
			when 12 => return "11110" & "11110" & "00001";
			when 13 => return "11110" & "11110" & "01001";
			when 14 => return "11110" & "11110" & "11110";
			when others => return "000000000000000";
		end case;
	end function color_lut;

begin

	-- use DCM to genereate clock signals
	dcm_inst: BUFG_CLKDV_SUBM port map(
		CLK_IN => clk_pad,
		RST    => rst,
		CLK1X  => clk,
		CLK_DIV => clk_2,
		LOCK   => LOCK);

	-- the PS2-mouse controller
	mouse_inst: mouse port map(
		clk => clk,
		rst => rst,
		interrupt => mouse_interrupt,
		data      => mouse_data,
		mouse_clk_I => mouse_clk_I,
		mouse_clk_O => mouse_clk_O,
		mouse_clk_T => mouse_clk_T,
		mouse_data_I => mouse_data_I,
		mouse_data_O => mouse_data_O,
		mouse_data_T => mouse_data_T);


	ms_clk  <= mouse_clk_O when mouse_clk_T = '0' else 'Z';
	ms_data <= mouse_data_O when mouse_data_T = '0' else 'Z';
	mouse_clk_I <= ms_clk;
	mouse_data_I <= ms_data;

-- this stuff is also working
--	iobuf_ms_clk : IOBUF port map(
--		I => mouse_clk_O,
--		IO => ms_clk,
--		O => mouse_clk_I,
--		T => mouse_clk_T
--	);

--	iobuf_ms_data : IOBUF port map(
--		I => mouse_data_O,
--		IO => ms_data,
--		O => mouse_data_I,
--		T => mouse_data_T
--	);

	-- the VGA timing component
	vga_timing_inst: vga_timing port map(
		clk   => clk_2,
		rst   => rst,
		hsync => hsync_int,
		vsync => vsync_int,
		blank => blank,
		x     => x,
		y     => y);

	hsync <= hsync_int;
	vsync <= vsync_int;

	-- the 8 pixel calculating units
	gen_loop: for i in 0 to 7 generate
		calc_inst: pixel_calc port map(
			clk    => clk,
			rst    => rst,
			start  => start,
			c_x    => c_x(i),
			c_y    => c_y,
			result => result(i));
	end generate;

	
	master_ctrl: process(clk, rst) is
	begin
		if rst = '1' then
		
			clock_counter     <= (others => '0');
			master_ctrl_state <= IDLE;
			for i in 0 to 7 loop
				c_x(i)        <= (others => '0');
				result_reg(i) <= (others => '0');
			end loop;
			c_y   <= to_signed(START_Y, 18);
			start <= '0';
			
		elsif clk'event and clk = '1' then

			start       <= '0'; -- set to '1' for one clock_cycle only
			
			if hsync_int = '0' and hsync_int_t = '1' then
				-- reset c_x, increment c_y
				for i in 0 to 7 loop
					c_x(i) <= start_x_sig + x_offset(i);
				end loop;
				c_y <= c_y - y_offset;
			end if;
			
			if vsync_int = '0' and vsync_int_t = '1' then
				-- reset c_y
				c_y <= start_y_sig;
			end if;

			case master_ctrl_state is
				when IDLE => 
					if blank = '0' then
						master_ctrl_state <= RUNNING;
						start             <= '1';
					end if;

				when RUNNING => 
					clock_counter <= clock_counter + 1;
					if clock_counter = 15 then
						clock_counter <= (others => '0');
						start <= '1';
						
						-- Increment by x_offset*8 each 8 pixel block
						for i in 0 to 7 loop
							c_x(i) <= c_x(i) + (x_offset(0)(14 downto 0) & "000");
							result_reg(i) <= result(i);
						end loop;
						if blank = '1' then
							master_ctrl_state <= IDLE;
							start <= '0';
						end if;
					end if;
			end case;
		end if;
	end process master_ctrl;

	vga_output: process(clk_2, rst) is
	begin
		if rst = '1' then
			rgb(14 downto 0) <= (others => '0');
		elsif clk_2'event and clk_2 = '1' then
			if blank_delay(17) = '0' then
				rgb(14 downto 0) <= color_lut(result_reg(to_integer(unsigned(x(2 downto 0))-1 )));
			else
				rgb(14 downto 0) <= (others => '0');
			end if;
		end if;
	end process vga_output;
	

	misc_stuff: process(clk, rst) is
	begin
		if rst = '1' then
			hsync_int_t <= '0';
			vsync_int_t <= '0';
			-- do not reset blank_delay! synthesis wouldn't use SRL primitives then.
		elsif clk'event and clk = '1' then

			vsync_int_t <= vsync_int;
			hsync_int_t <= hsync_int;
			blank_delay(0) <= blank;
			blank_delay(17 downto 1) <= blank_delay(16 downto 0);
		end if;
	end process misc_stuff;


	zoom_control: process(clk, rst) is
	begin
		if rst = '1' then
			start_x_sig <= to_signed(START_X, 18);
			start_y_sig <= to_signed(START_Y, 18);
			index     <= 0;
			x_offset(0) <= to_signed(127, 18);
			for i in 1 to 7 loop
				x_offset(i) <= (others => '0');
			end loop;
			--y_offset <= (others => '0');
			y_offset <= to_signed(127, 18);
			state_zoom_auto <= '0';
			zoom_in <= '0';
		elsif clk = '1' and clk'event then
		
			if mouse_interrupt = '1' then
				start_x_sig <= start_x_sig + signed((mouse_data(15 downto 7) & "0000000"));
				start_y_sig <= start_y_sig + signed((mouse_data(23 downto 16) & "0000000"));
				if mouse_data(0) = '1' then -- right button : zoom out
					x_offset(0) <= x_offset(0) - 2;
					y_offset <= y_offset - 2;
				end if;
				if mouse_data(1) = '1' then -- left button: zoom in
					x_offset(0) <= x_offset(0) + 2;
					y_offset <= y_offset + 2;
				end if;
				if mouse_data(2) = '1' then -- middle button : zoom auto
					state_zoom_auto <= not state_zoom_auto;
					if state_zoom_auto = '0' then -- only reset when we enter the auto zoom mode
						x_offset (0) <= (others => '0'); -- reset the offsets to prevent running out of
						y_offset     <= (others => '0'); -- the automatic zooming borders!
						zoom_in      <= '0';
					end if;
				end if;
			end if;

			if vsync_int = '0' and vsync_int_t = '1' then
				index <= 1;

				if state_zoom_auto = '1' then
					if zoom_in = '1' then
						x_offset(0) <= x_offset(0) - 1;
						y_offset    <= y_offset - 1;
						if x_offset(0) = 1 then
							zoom_in <= '0';
						end if;
					elsif zoom_in = '0' then -- zoom out
						x_offset(0) <= x_offset(0) + 1;
						y_offset    <= y_offset + 1;
						if x_offset(0)(7) = '1' then -- zoom factor > 127?
							zoom_in <= '1';
						end if;
					end if;
				end if;

			end if;
			-- hardware loop index of (0..7) to initialise x_offset vector
			if index /= 0 then
				x_offset(index) <= x_offset(0) + x_offset(index-1);
				if index = 7 then
					index <= 0;
				else
					index <= index + 1;
				end if;
			end if;
		end if;
	end process;


end architecture rtl;

