-- Project: Fractal for Spartan-3
-- Author : Matthias Alles
-- Mail   : alles(at) rhrk.uni-kk.de
--
-- This module initializes a PS2 mouse and gives data when available to the environment.
-- Initialization takes 100us, adapt the constant when using with different clock frequency.
-- This module was used at 49,152MHz

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;


entity mouse is
	port
	(
      clk          : in  std_logic;
	  rst          : in  std_logic;
	  interrupt    : out std_logic;
	  data         : out std_logic_vector(31 downto 0);
      mouse_clk_I  : in  std_logic;
      mouse_clk_O  : out std_logic;
      mouse_clk_T  : out std_logic;
      mouse_data_I : in  std_logic;
      mouse_data_O : out std_logic;
      mouse_data_T : out std_logic
	);
end entity mouse;



architecture rtl of mouse is

	-- this register contains the p2-data we received and is used by the driver
	--type register32 is array(0 to C_NUM_CE-1) of std_logic_vector(0 to 31);
	signal ps2_data_reg: std_logic_vector(31 downto 0);

	-- mouse init
	type mouse_init_state is (CLEAR_CLK, RELEASE_CLK, SEND, WAIT_ACK, RECEIVE, DONE);

	signal mouse_state: mouse_init_state;
	signal dir: std_logic;
	constant mouse_init: std_logic_vector(10 downto 0) :="11011110100"; 
	signal ptr: integer range 0 to 10;
	signal mouse_init_vec: std_logic_vector(32 downto 0);

	-- handling of ps2-keyboard
	signal ms_strobe   : std_logic;
	signal ms_data     : std_logic_vector(32 downto 0);

	signal ms_clk_t: std_logic;
	signal ms_clk_t2: std_logic;

	signal input_vector_mouse: std_logic_vector(32 downto 0);

	signal mouse_init_counter: unsigned(13 downto 0);
	signal init_done: std_logic;

begin

	ms_strobe <= not input_vector_mouse(0);
	ms_data   <= input_vector_mouse;
	data      <= ps2_data_reg;


	handle_mouse: process(clk) is
	begin
		if clk = '1' and clk'event then
			if rst = '1' then

				input_vector_mouse <= (others => '1');
				ms_clk_t     <= '0';
				ms_clk_t2    <= '0';
			else
				ms_clk_t  <= mouse_clk_i;
				ms_clk_t2 <= ms_clk_t;
				if ms_clk_t2 = '1' and ms_clk_t = '0' and init_done = '1' then  -- falling edge detected

					input_vector_mouse(31 downto 0) <= input_vector_mouse(32 downto 1);
					input_vector_mouse(32)          <= mouse_data_i;
				end if;

				if input_vector_mouse(0) = '0' then
					input_vector_mouse(32 downto 0) <= (others => '1');
				end if;
			end if;
		end if;
	end process handle_mouse;



	mouse_top: process(clk) is 
	begin

		if clk = '1' and clk'event then
			if rst = '1' then

				interrupt    <= '0';
				ps2_data_reg <= (others => '0');

			else
			
				interrupt <= '0';
				if ms_strobe = '1' then
					ps2_data_reg <= "00000000" &  ms_data(30 downto 23) & ms_data(19 downto 12) & ms_data(8 downto 1);
					interrupt <= '1';
				end if;
			end if;
		end if;
	end process mouse_top;


	init_mouse: process(clk) is
	begin
		if clk = '1' and clk'event then
			if rst = '1' then
				mouse_data_t <= '0'; --drive when starting;
				mouse_data_o <= '0';
				mouse_state  <= CLEAR_CLK;
				mouse_init_vec <= (others => '1');
				ptr          <= 0;
				init_done    <= '0';
				mouse_clk_t  <= '0';
				mouse_clk_o  <= '0';
				mouse_init_counter <= (others => '0');
			else

				case mouse_state is
					when CLEAR_CLK  =>
						mouse_clk_t <= '0';  -- drive clock signal for 100us low
						mouse_clk_o <= '0';
						mouse_data_t <= '1';  -- don't drive data line
						mouse_init_counter <= mouse_init_counter + 1;
						if mouse_init_counter = 4915 then -- wait 100us
							mouse_state <= RELEASE_CLK;
							mouse_init_counter <= (others => '0');
							mouse_data_o <= '0';
							mouse_data_t <= '0'; -- drive data low too 
						end if;

					when RELEASE_CLK =>
						mouse_init_counter <= mouse_init_counter + 1; --wait some time
						if mouse_init_counter = 150 then
							mouse_clk_t      <= '1'; --release clock signal and wait for falling edge
							mouse_state      <= SEND;
						end if;
					
					when SEND =>
						if ms_clk_t2 = '1' and ms_clk_t = '0' then  -- falling edge detected
							mouse_data_o   <= mouse_init(ptr);
							ptr            <= ptr + 1;
							if ptr = 9 then
								mouse_data_t <= '1';  --release data_line
								mouse_state <= WAIT_ACK;
								ptr <= 0;
							end if;
						end if;

					when WAIT_ACK =>
						if ms_clk_t2 = '1' and ms_clk_t = '0' then  -- falling edge detected
							if mouse_data_i = '0' then  -- acknowledge by mouse
								mouse_state <= RECEIVE;
							end if;
						end if;

					when RECEIVE =>
						if ms_clk_t2 = '1' and ms_clk_t = '0' then  -- falling edge detected, we receive with falling edge
							mouse_init_vec(9 downto 0) <= mouse_init_vec(10 downto 1);
							mouse_init_vec(10) <= mouse_data_I;
						end if;
						if mouse_init_vec(0) = '0' then
							mouse_state <= DONE;
							mouse_init_vec(32 downto 0) <= (others => '1');
						end if;             

					when DONE   =>
						init_done <= '1';

					when others =>
				end case; 
			end if;
		end if;
	end process init_mouse;

end architecture rtl;
